use tracing::debug;

use ubisync_lib::peer::Peer;
use ubisync_lib::types::{ContentUpdateStrategy, Family, PeerId};

use ubisync_lib::messages::{Message, MessageContent};

use crate::comm::conflict_resolution::merge_element_contents;
use crate::state::CommState;

pub fn handle(state: &CommState, peer: &PeerId, message: Message) {
    debug!(
        "Received message.\nFrom: {:?} (dest: {:?})\nTo: {:?} (dest: {:?})\nMessage: {message:?}",
        peer.b32_addr_nocache(),
        peer,
        state.own_peer_id().unwrap().b32_addr_nocache(),
        state.own_peer_id().unwrap()
    );
    match message.content() {
        MessageContent::Hello { peer_name } => {
            state
                .set_peer(Peer::new(peer.to_owned(), Some(peer_name.to_string())))
                .expect("State failed");
        }
        MessageContent::JoinFamily => state.request_family_join(peer.to_owned()),
        MessageContent::AddedToFamily { family } => {
            if state.has_family_join_request(peer.to_owned()) {
                debug!("Own join request was accepted, setting family");
                state
                    .set_own_family(family.to_owned())
                    .expect("State failed");
                
                debug!("New own family: {:?}", state.get_family_of_peer(state.own_peer_id().unwrap()))
            }
            else {
                debug!("Got AddedToFamily message, but no family join request was found")
            }
        }
        MessageContent::LeaveFamily => state.remove_peer_from_family(peer.to_owned()),
        MessageContent::CreateElement { id, content, pot } => {
            state
                .add_received_element(
                    id.to_owned(),
                    content.to_owned(),
                    ContentUpdateStrategy::Overwrite,
                    Some(message.id().to_owned()),
                    pot.to_owned(),
                )
                .expect("State failed");
        }
        MessageContent::UpdateElement { id, content } => {
            if let Some(new_content) = merge_element_contents(
                state,
                id.clone(),
                content.to_owned(),
                message.id().to_owned(),
            ) {
                state
                    .update_element_content(
                        id.to_owned(),
                        new_content.to_owned(),
                        message.id().to_owned(),
                    )
                    .expect("State failed");
            }
        }
        MessageContent::RemoveElement { id } => {
            state.remove_element(id.to_owned()).expect("State failed");
        }
        MessageContent::AddPot { id, app_type } => {
            state
                .add_pot(id.to_owned(), app_type.to_string())
                .expect("State failed");
        }
    }
}
