use std::{future::Future, sync::Arc};

use anyhow::bail;
use api::{v0::app::App, Api, ApiBuilder};
use comm::CommHandle;
use config::Config;
use i2p::net::I2pSocketAddr;
use node_events::UbisyncNodeEvent;
use state::{ApiState, CommState, State};

use ubisync_lib::{
    messages::{Message, MessageContent},
    peer::Peer,
    types::{AppId, Family, PeerId, PotId},
};

pub mod api;
pub mod comm;
pub mod config;
pub mod node_events;
pub mod state;

pub struct Ubisync {
    comm_handle: Arc<CommHandle>,
    state_handle: Arc<State>,
    api: Arc<Api>,
}

impl Ubisync {
    pub async fn new(config: &Config) -> anyhow::Result<Self> {
        let state = State::new(config.database_location.clone()).await?;
        let comm_handle = Arc::new(CommHandle::new(CommState::new(state.clone()), config)?);
        state.set_comm_handle(comm_handle.clone());

        let api = Arc::new(
            ApiBuilder::from(config.api_config.clone())
                .build(ApiState::new(state.clone(), &config.jwt_secret))
                .await,
        );

        comm_handle.run().await;
        Ok(Ubisync {
            comm_handle: comm_handle,
            state_handle: state,
            api: api,
        })
    }

    pub fn set_node_event_callback<CallbackFunction>(
        &self,
        cb: CallbackFunction,
        node: Arc<Ubisync>,
    ) where
        CallbackFunction: Fn(UbisyncNodeEvent, Arc<Ubisync>) + Send + Sync + 'static,
    {
        self.state_handle.set_node_event_callback(cb, node);
    }

    pub fn api(&self) -> Arc<Api> {
        self.api.clone()
    }

    pub fn add_peer(&self, p: impl TryInto<Peer, Error = anyhow::Error>) -> anyhow::Result<()> {
        match p.try_into() {
            Ok(peer) => self.state_handle.add_peer(peer),
            Err(e) => bail!(e),
        }
    }

    pub fn add_peer_from_id(&self, id: PeerId) -> anyhow::Result<()> {
        // TODO: resolve peer's name before setting
        self.state_handle.add_peer(Peer::new(id, None))
    }

    pub fn add_family_member_from_id(&self, id: PeerId) -> anyhow::Result<()> {
        self.add_peer_from_id(id.clone())?;
        self.state_handle.add_family_member(id)
    }

    pub async fn request_family_join(&self, peer: PeerId) -> anyhow::Result<()> {
        self.state_handle.add_family_join_request(peer.clone());
        self.comm_handle
            .send(peer.addr_ref(), Message::new(MessageContent::JoinFamily))
            .await
    }

    pub fn accept_family_join(&self, peer: PeerId) -> anyhow::Result<()> {
        self.add_family_member_from_id(peer.clone())?;
        let family = self.get_family()?;
        tokio::spawn({
            let ch = self.comm_handle.clone();
            async move {
                ch.send(
                    peer.addr_ref(),
                    Message::new(MessageContent::AddedToFamily { family }),
                )
                .await
                .expect("Could not send family join confirmation to peer");
            }
        });
        Ok(())
    }

    pub fn get_family(&self) -> anyhow::Result<Family> {
        self.state_handle
            .get_family_of(self.comm_handle.own_peer_id()?)
    }

    pub fn get_family_of(&self, peer: PeerId) -> anyhow::Result<Family> {
        self.state_handle.get_family_of(peer)
    }

    pub fn get_apps(&self) -> Vec<App> {
        self.state_handle.get_apps().unwrap_or(vec![])
    }

    pub fn get_peers(&self) -> Vec<Peer> {
        self.state_handle.get_peers().unwrap_or(vec![])
    }

    pub fn get_destination(&self) -> I2pSocketAddr {
        self.comm_handle.i2p_address()
    }

    pub fn add_pot_member(&self, pot: PotId, app: AppId) -> anyhow::Result<()> {
        self.state_handle.add_pot_member(pot, app)
    }
}

#[cfg(test)]
mod tests {
    use std::{sync::Arc, time::Duration};

    use tracing::{debug, Level};

    use crate::{config::Config, node_events::UbisyncNodeEvent, Ubisync};

    #[tokio::test(flavor = "multi_thread")]
    async fn join_and_leave_family() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        // Two nodes need to bind to different ports
        let mut c2 = Config::default();
        c2.api_config.port = Some(9982);
        let ubi1 = Arc::new(Ubisync::new(&Config::default()).await.unwrap());
        let ubi2 = Arc::new(Ubisync::new(&c2).await.unwrap());

        ubi1.set_node_event_callback(
            move |ev, node| {
                if let UbisyncNodeEvent::FamilyJoinRequest { joiner } = ev {
                    debug!("Received join request, make member join");
                    node.accept_family_join(joiner).unwrap();
                }
            },
            ubi1.clone(),
        );

        ubi2.request_family_join(ubi1.get_destination().into())
            .await
            .unwrap();

        tokio::time::sleep(Duration::from_millis(5000)).await;

        assert_eq!(ubi1.get_family().unwrap(), ubi2.get_family().unwrap());
        std::process::exit(0);
    }
}
