use anyhow::{anyhow, Error};
use bonsaidb::core::schema::{Collection, SerializedCollection};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use ubisync_lib::{
    peer::Peer,
    types::{ElementId, PeerId},
};

use crate::state::database::{as_key::AsKey, StateDB};

use super::peer_families::DbPeerFamily;

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "peers", views = [])]
pub(super) struct DbPeer {
    #[natural_id]
    pub(super) id: AsKey<PeerId>,
    pub(super) name: Option<String>,
}

impl From<DbPeer> for Peer {
    fn from(value: DbPeer) -> Self {
        Peer::new((*value.id).clone(), value.name)
    }
}

impl StateDB {
    pub fn add_peer(&self, peer: Peer) -> anyhow::Result<()> {
        if self.get_peer(peer.id())?.is_some() {
            Err(Error::msg("Peer already exists"))
        } else {
            DbPeer::push(
                DbPeer {
                    id: AsKey::new(peer.id()),
                    name: peer.name(),
                },
                &self.db,
            )
            .map(|_| ())
            .map_err(|e| anyhow!(e))
        }
    }

    pub fn get_peer(&self, id: PeerId) -> anyhow::Result<Option<Peer>> {
        DbPeer::get(&AsKey::new(id), &self.db)
            .map(|doc| doc.map(|peer| Peer::new((*peer.contents.id).clone(), peer.contents.name)))
            .map_err(|e| anyhow!(e))
    }

    pub fn get_all_peers(&self) -> anyhow::Result<Vec<Peer>> {
        DbPeer::all(&self.db)
            .query()
            .map(|peers| {
                peers
                    .iter()
                    .map(|p| p.contents.clone().into())
                    .collect_vec()
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn set_peer_name(&self, id: PeerId, name: &Option<String>) -> anyhow::Result<()> {
        DbPeer::get(&AsKey::new(id), &self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Peer not found"))?
            .modify(&self.db, |doc| doc.contents.name = name.clone())
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::{peer::Peer, types::PeerId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);

        let peer = Peer::new(PeerId::default(), Some("Peer name".to_string()));
        db.add_peer(peer.clone()).unwrap();

        let retrieved_peer = db.get_peer(peer.id()).unwrap();

        assert_eq!(Some(peer), retrieved_peer)
    }

    #[test]
    fn get_all() {
        let db = StateDB::init(None);

        let peer = Peer::new(PeerId::default(), Some("Peer name".to_string()));
        db.add_peer(peer.clone()).unwrap();

        let all_peers = db.get_all_peers().unwrap();
        assert_eq!(all_peers, vec![peer]);
    }

    #[test]
    fn set_peer_name() {
        let db = StateDB::init(None);

        let peer = Peer::new(PeerId::default(), Some("Peer name".to_string()));
        db.add_peer(peer.clone()).unwrap();
        db.set_peer_name(peer.id().clone(), &Some("New peer name".to_string()))
            .unwrap();

        let retrieved_peer = db.get_peer(peer.id()).unwrap();

        assert_eq!(
            Some(Peer::new(peer.id(), Some("New peer name".to_string()))),
            retrieved_peer
        )
    }
}
