use anyhow::{Error, anyhow};
use bonsaidb::core::schema::{Collection, SerializedCollection};
use serde::{Deserialize, Serialize};
use ubisync_lib::types::{Pot, PotId};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "pots", views = [])]
pub(super) struct DbPot {
    #[natural_id]
    pub(super) id: AsKey<PotId>,
    pub(super) app_type: String,
}

impl From<DbPot> for Pot {
    fn from(value: DbPot) -> Self {
        Pot {
            id: (*value.id).clone(),
            app_type: value.app_type,
        }
    }
}

impl StateDB {
    pub fn add_pot(&self, id: PotId, app_type: String) -> anyhow::Result<()> {
        if self.get_pot(id.clone())?.is_some() {
            Err(Error::msg("Pot already exists"))
        } else {
            DbPot::push(
                DbPot {
                    id: AsKey::new(id),
                    app_type,
                },
                &self.db,
            )
            .map(|_| ())
            .map_err(|e| anyhow!(e))
        }
    }

    pub fn get_pot(&self, id: PotId) -> anyhow::Result<Option<Pot>> {
        DbPot::get(&AsKey::new(id), &self.db)
            .map(|pot_opt| pot_opt.map(|pot| pot.contents.into()))
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{Pot, PotId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let pot_id = PotId::new();
        db.add_pot(pot_id.clone(), "app_type".to_string()).unwrap();

        let retrieved_pot = db.get_pot(pot_id.clone()).unwrap();
        assert_eq!(
            retrieved_pot,
            Some(Pot {
                id: pot_id,
                app_type: "app_type".to_string()
            })
        )
    }
}
