use anyhow::anyhow;
use bonsaidb::{
    core::keyvalue::KeyValue,
    local::{
        config::{Builder, StorageConfiguration},
        Database as BonsaiDb,
    },
};
use tracing::debug;
use ubisync_lib::types::PeerId;
use uuid::Uuid;

use self::collections::UbisyncSchema;

mod as_key;
mod collections;

pub struct StateDB {
    db: BonsaiDb,
}

impl StateDB {
    pub fn init(path: Option<String>) -> Self {
        let storage_conf = match path {
            Some(p) => StorageConfiguration::new(p),
            None => StorageConfiguration::new(format!("/tmp/{}", Uuid::new_v4())),
            // None => StorageConfiguration::default().memory_only()
        };
        StateDB {
            db: BonsaiDb::open::<UbisyncSchema>(storage_conf).unwrap(),
        }
    }

    pub fn add_family_join_request(&self, peer: PeerId) {
        self.db.set_key(peer.to_string(), &"").execute();
        debug!(
            "Added join request: {:?}",
            self.db.get_key(peer.to_string()).query()
        )
    }

    pub fn has_family_join_request(&self, peer: PeerId) -> anyhow::Result<bool> {
        Ok(self
            .db
            .get_key(peer.to_string())
            .query()
            .map_err(|e| anyhow!(e))?
            .is_some())
    }

    pub fn remove_family_join_request(&self, peer: PeerId) -> anyhow::Result<()> {
        self.db
            .delete_key(peer.to_string())
            .map(|_| ())
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::PeerId;

    use super::StateDB;

    #[test]
    fn family_join_requests() {
        let db = StateDB::init(None);
        let peer = PeerId::default();
        db.add_family_join_request(peer.clone());

        assert!(db.has_family_join_request(peer).unwrap());
    }
}
